# Copyright (c) 2019-2022, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from forge.command.forge import ForgeCommand
from forge.printer.instance import InstancePrinter
from ngcbpc.api.configuration import Configuration
from ngcbpc.command.args_validation import check_add_args_columns, check_valid_columns
from ngcbpc.command.clicommand import CLICommand
from ngcbpc.constants import CONFIG_TYPE, DISABLE_TYPE, STAGING_ENV
from ngcbpc.errors import InvalidArgumentError
from ngcbpc.util.utils import get_columns_help, get_environ_tag, has_org_role
from ngccli.modules.client import Client


class InstanceCommand(ForgeCommand):

    CMD_NAME = "instance"
    HELP = "Instance Commands"
    DESC = "Instance Commands"

    def __init__(self, parser):
        super().__init__(parser)
        self.parser = parser
        self.config = Configuration()
        self.client = Client(api_key=self.config.app_key, base_url=self.config.base_url)
        self.api = self.client.forge.instance
        self.printer = InstancePrinter()

    LIST_HELP = "List instances."
    columns_dict = {
        "name": "Name",
        "allocationId": "Allocation Id",
        "allocationName": "Allocation Name",
        "tenantId": "Tenant Id",
        "tenantName": "Tenant Name",
        "infrastructureProviderId": "Infrastructure Provider Id",
        "infrastructureProviderName": "Infrastructure Provider Name",
        "siteId": "Site Id",
        "siteName": "Site Name",
        "instanceTypeId": "Instance Type Id",
        "instanceTypeName": "Instance Type Name",
        "vpcId": "VPC Id",
        "vpcName": "VPC Name",
        "machineId": "Machine Id",
        "machineName": "Machine Name",
        "operatingSystemId": "Operating System Id",
        "operatingSystemName": "Operating System Name",
        "ipxeScript": "IPXE Script",
        "userdata": "User Data",
        "status": "Status",
        "created": "Created",
        "updated": "Updated",
    }
    columns_default = ("id", "Id")
    columns_help = get_columns_help(columns_dict, columns_default)
    status_enum = ["Pending", "Provisioning", "Configuring", "Ready", "Rebooting", "Terminating", "Error"]

    @CLICommand.arguments(
        "target",
        metavar="<target>",
        help=(
            "Filter by matches across all instances. Input will be matched against name, description and status fields."
        ),
        type=str,
        nargs="?",
        default=None,
    )
    @CLICommand.arguments(
        "--status",
        metavar="<status>",
        help=f"Filter by status. Choices are: {', '.join(status_enum)}",
        type=str,
        default=None,
        choices=status_enum,
    )
    @CLICommand.arguments(
        "--column",
        metavar="<column>",
        help=columns_help,
        default=None,
        action="append",
        type=lambda value, columns_dict=columns_dict: check_valid_columns(value, columns_dict),
    )
    @CLICommand.arguments("--site", metavar="<site>", help="Filter by site id.", type=str, required=True)
    @CLICommand.arguments("--vpc", metavar="<vpc>", help="Filter by vpc id.", type=str)
    @CLICommand.command(help=LIST_HELP, description=LIST_HELP)
    def list(self, args):
        """List Instances."""
        self.config.validate_configuration()
        org_name = args.org or self.config.org_name
        team_name = args.team or self.config.team_name
        provider = None
        tenant = None
        user_resp = self.client.users.user_who(org_name)
        if has_org_role(user_resp, org_name, ["FORGE_PROVIDER_ADMIN"]):
            provider_info, _ = self.client.forge.provider.info(org_name, team_name)
            provider = provider_info.get("id", "")
        elif has_org_role(user_resp, org_name, ["FORGE_TENANT_ADMIN"]):
            tenant_info, _ = self.client.forge.tenant.info(org_name, team_name)
            tenant = tenant_info.get("id", "")
        resp = self.api.list(org_name, team_name, args.site, args.vpc, provider, tenant, args.target, args.status)
        check_add_args_columns(args.column, InstanceCommand.columns_default)
        self.printer.print_list(resp, args.column)

    INFO_HELP = "Instance information."

    @CLICommand.arguments("instance", metavar="<instance>", help="Instance id.", type=str)
    @CLICommand.arguments("--status-history", help="Show status history.", action="store_true")
    @CLICommand.command(help=INFO_HELP, description=INFO_HELP)
    def info(self, args):
        """Instance info."""
        self.config.validate_configuration()
        org_name = args.org or self.config.org_name
        team_name = args.team or self.config.team_name
        resp = self.api.info(org_name, team_name, args.instance)
        self.printer.print_info(resp, args.status_history)

    CREATE_HELP = "Create instance."
    IB_TYPE = CONFIG_TYPE if (get_environ_tag() <= STAGING_ENV) else DISABLE_TYPE

    @CLICommand.arguments("name", metavar="<name>", help="Instance name.", type=str)
    @CLICommand.arguments(
        "--instance-type", metavar="<instance_type>", help="Specify instance type id.", type=str, required=True
    )
    @CLICommand.arguments("--vpc", metavar="<vpc>", help="Specify VPC id.", type=str, required=True)
    @CLICommand.arguments(
        "--operating-system", metavar="<operating_system>", help="Specify operating system id.", type=str
    )
    @CLICommand.arguments("--user-data", metavar="<user_data>", help="Specify user data.", type=str)
    @CLICommand.arguments(
        "--ipxe-script",
        metavar="<ipxe_script>",
        help="Override iPXE script specified in OS, must be specified if OS is not specified",
        type=str,
    )
    @CLICommand.arguments(
        "--always-boot-custom-ipxe",
        help="Always reboots the instance using iPXE script specified by OS or instance. OS must be of type iPXE.",
        action="store_true",
    )
    @CLICommand.arguments(
        "--interface",
        metavar="<subnet:physical>",
        help=(
            "Specify subnet id and is physical <true|false>. Only one is attached over physical interface."
            "First one is attached over physical interface if not specified. Multiple interface arguments are allowed."
        ),
        type=str,
        action="append",
        required=True,
    )
    @CLICommand.arguments(
        "--infiniband-interface",
        metavar="<partition:device:vendor:device_instance:physical:virtual>",
        help=(
            "Specify partition id, device name, vendor name (optional), device instance, is physical <true|false> "
            "and virtual function id (must be specified if physical is false). "
            "Multiple infiniband interface arguments are allowed."
        ),
        type=str,
        action="append",
        feature_tag=IB_TYPE,
    )
    @CLICommand.arguments(
        "--ssh-key-group",
        metavar="ssh_key_group",
        help="Specify SSH Key Group IDs that will provide Serial over LAN access.",
        type=str,
        action="append",
    )
    @CLICommand.command(help=CREATE_HELP, description=CREATE_HELP)
    def create(self, args):
        """Create Instance."""
        self.config.validate_configuration()
        org_name = args.org or self.config.org_name
        team_name = args.team or self.config.team_name
        tenant = None
        tenant_info, _ = self.client.forge.tenant.info(org_name, team_name)
        tenant = tenant_info.get("id", "")
        interfaces = []
        for it in args.interface or []:
            try:
                il = it.split(":")
                il += [None] * (2 - len(il))
                if il[1] is not None:
                    if il[1] in ("true", "t", "yes", "y", "1"):
                        il[1] = True
                    elif il[1] in ("false", "f", "no", "n", "0"):
                        il[1] = False
                    else:
                        raise InvalidArgumentError(f"interface: {it}") from None
                    interfaces.append({"subnetId": il[0], "isPhysical": il[1]})
                else:
                    interfaces.append({"subnetId": il[0]})
            except (ValueError, TypeError, AttributeError, IndexError):
                raise InvalidArgumentError(f"interface: {it}") from None
        infiniband_interfaces = []
        for inb in args.infiniband_interface or []:
            try:
                inl = inb.split(":")
                inl += [None] * (6 - len(inl))
                if inl[4] is not None:
                    if inl[4] in ("true", "t", "yes", "y", "1"):
                        inl[4] = True
                    elif inl[4] in ("false", "f", "no", "n", "0"):
                        inl[4] = False
                    else:
                        raise InvalidArgumentError(f"infiniband interface: {inb}") from None
                inl[3] = int(inl[3]) if inl[3] is not None else inl[3]
                inl[5] = int(inl[5]) if inl[5] is not None else inl[5]
                infiniband_interfaces.append(
                    {
                        "partitionId": inl[0],
                        "device": inl[1],
                        "vendor": inl[2],
                        "deviceInstance": inl[3],
                        "isPhysical": inl[4],
                        "virtualFunctionId": inl[5],
                    }
                )
            except (ValueError, TypeError, AttributeError, IndexError):
                raise InvalidArgumentError(f"infiniband interface: {inb}") from None
        resp = self.api.create(
            org_name,
            team_name,
            args.name,
            tenant,
            args.instance_type,
            args.vpc,
            args.operating_system,
            args.user_data,
            args.ipxe_script,
            args.always_boot_custom_ipxe,
            interfaces,
            infiniband_interfaces,
            args.ssh_key_group,
        )
        self.printer.print_info(resp)

    UPDATE_HELP = "Update instance."

    @CLICommand.arguments("instance", metavar="<instance>", help="Instance id.", type=str)
    @CLICommand.arguments("--name", metavar="<name>", help="Specify instance name.", type=str)
    @CLICommand.arguments("--reboot", help="Trigger instance reboot.", action="store_true")
    @CLICommand.arguments(
        "--reboot-custom-ipxe",
        help="Reboot using the custom iPXE specified by OS, is ignored if instance has alwaysBootWithCustomIpxe.",
        action="store_true",
    )
    @CLICommand.command(help=UPDATE_HELP, description=UPDATE_HELP)
    def update(self, args):
        """Update Instance."""
        self.config.validate_configuration()
        org_name = args.org or self.config.org_name
        team_name = args.team or self.config.team_name
        resp = self.api.update(org_name, team_name, args.instance, args.name, args.reboot, args.reboot_custom_ipxe)
        self.printer.print_info(resp)

    REMOVE_HELP = "Remove instance."

    @CLICommand.arguments("instance", metavar="<instance>", help="Instance id.", type=str)
    @CLICommand.command(help=REMOVE_HELP, description=REMOVE_HELP)
    def remove(self, args):
        """Remove Instance."""
        self.config.validate_configuration()
        org_name = args.org or self.config.org_name
        team_name = args.team or self.config.team_name
        resp = self.api.remove(org_name, team_name, args.instance)
        self.printer.print_ok(f"{resp}")

    LIST_INTERFACE_HELP = "List all interfaces for the instance."

    @CLICommand.arguments("instance", metavar="<instance>", help="Instance id.", type=str)
    @CLICommand.command(name="list-interface", help=LIST_INTERFACE_HELP, description=LIST_INTERFACE_HELP)
    def list_interface(self, args):
        """List Instance Interfaces."""
        self.config.validate_configuration()
        org_name = args.org or self.config.org_name
        team_name = args.team or self.config.team_name
        resp = self.api.list_interface(org_name, team_name, args.instance)
        self.printer.print_list_interface(resp)
